#define BOOST_ALL_NO_LIB
#include <windows.h>
#include "avisynth.h"
#include <deque>
#include <utility>

#include <boost/thread.hpp>
#include <boost/bind.hpp>

using boost::thread;
using boost::recursive_mutex;
using std::deque;
using boost::condition_variable_any;
using boost::unique_lock;

class SoraThread : public GenericVideoFilter
{
    struct BufferedFrame
    {
        int nFrame; //֡
        int nLifeTimeLeft; //ʼֵڻ֡ÿһξͼһС0ʱ
        PVideoFrame frame; //֡
        
        const BufferedFrame& operator = (const BufferedFrame& bf)
        {
            nFrame = bf.nFrame;
            nLifeTimeLeft = bf.nLifeTimeLeft;
            frame = bf.frame;
            return *this;
        }
        
        BufferedFrame(const BufferedFrame& a)
        {
            *this = a;
        }
        
        BufferedFrame() { }
    };
    
    private:
        //֡
        int nFrameCount;
        
        //棬֡ݴ
        int nCacheCount;
        deque<BufferedFrame> frameCache;
        recursive_mutex frameCacheMutex;
        
        //壬û֡ݴԺƶ
        int nBufferCount;
        deque<BufferedFrame> frameBuffer;
        recursive_mutex frameBufferMutex;
        
        //ʱ֪ͨ
        condition_variable_any frameBufferChangedEvent;
        
        //߳
        thread processThread;
        
        //һ֡ʲô
        int nNextFrameToRead;
        recursive_mutex nextFrameToReadMutex;
        
        //֪ͨ߳
        recursive_mutex startEventMutex;
        condition_variable_any startEvent;
        
        IScriptEnvironment* firstEnv; //һͽenv
    public:
        SoraThread(PClip _child, int bc, int cc)
            : GenericVideoFilter(_child)
            , nCacheCount(cc)
            , nBufferCount(bc)
        {
            nFrameCount = vi.num_frames;
            nNextFrameToRead = -1;
        }
        
        ~SoraThread()
        {
            if( processThread.get_id() != boost::thread::id() )
            {
                processThread.interrupt();
                processThread.join();
            }
        }
        
        PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env)
        {
            PVideoFrame frame;
            if( processThread.get_id() == boost::thread::id() )
            {
                firstEnv = env;
                processThread = thread(boost::bind(&SoraThread::threadFunction, this));
                nNextFrameToRead = n;
                WaitForFrame(n, frame);
                
                return frame;
            }
            if( SearchCache(n, frame) ) //ڻҵҪ֡
            {
                ;
            }
            else if( SearchBuffer(n, frame) ) //ڻҵҪ֡
            {
                ;
            }
            else //ûҵ趨һ֡Ժȴ
            {
                {
                    unique_lock<recursive_mutex> lg(nextFrameToReadMutex);
                    nNextFrameToRead = n;
                }
                frameBufferChangedEvent.notify_all();
                WaitForFrame(n, frame);
            }
            
            return frame;
        }
        
        bool SearchCache(int nFrame, PVideoFrame& frame) //һ棬ҵԺµڵһ
        {
            unique_lock<recursive_mutex> lg(frameCacheMutex);
            
            deque<BufferedFrame>::iterator i = frameCache.begin();
            
            while(i != frameCache.end())
            {
                if(i->nFrame == nFrame)
                {
                    frame = i->frame;
                    struct BufferedFrame bf = *i;
                    frameCache.erase(i);
                    CacheFrame(bf.nFrame, bf.frame);
                    return true;
                }
                ++i;
            }
            return false;
        }
        
        bool SearchBuffer(int nFrame, PVideoFrame& frame) //һ
        {
            unique_lock<recursive_mutex> lg(frameBufferMutex);
            
            deque<BufferedFrame>::iterator i = frameBuffer.begin();
            
            while(i != frameBuffer.end())
            {
                if(i->nFrame == nFrame)
                {
                    frame = i->frame;
                    CacheFrame(nFrame, frame);
                    i = frameBuffer.erase(i);
                    lg.unlock();
                    frameBufferChangedEvent.notify_all();
                    return true;
                }
                ++i;
            }
            return false;
        }
        
        void CacheFrame(int nFrame, PVideoFrame frame)
        {
            unique_lock<recursive_mutex> lg(frameCacheMutex);
            
            while(frameCache.size() == nCacheCount)
                frameCache.pop_back();
            
            struct BufferedFrame bf;
            bf.frame = frame;
            bf.nFrame = nFrame;
            frameCache.push_front(bf);
        }
        
        void BufferFrame(int nFrame, PVideoFrame frame)
        {
            unique_lock<recursive_mutex> lg(frameBufferMutex);
            
            while(frameBuffer.size() == nBufferCount) //˻ȴ
            {
                frameBufferChangedEvent.wait(lg);
            }
            
            ReduceBufferedFrameLifeTime();
            
            struct BufferedFrame bf;
            bf.frame = frame;
            bf.nFrame = nFrame;
            bf.nLifeTimeLeft = nBufferCount;
            frameBuffer.push_back(bf);
            
            lg.unlock();
            frameBufferChangedEvent.notify_all();
        }
        
        void ReduceBufferedFrameLifeTime()
        {
            unique_lock<recursive_mutex> lg(frameBufferMutex);
            
            deque<BufferedFrame>::iterator i = frameBuffer.begin();
            while(i != frameBuffer.end())
            {
                -- i->nLifeTimeLeft;
                if(i->nLifeTimeLeft == 0)
                    i = frameBuffer.erase(i);
                else
                    ++i;
            }
            
            lg.unlock();
            frameBufferChangedEvent.notify_all();
        }
        
        void WaitForFrame(int nFrame, PVideoFrame& frame)
        {
            unique_lock<recursive_mutex> lg(frameBufferMutex);
            
            while(SearchBuffer(nFrame, frame) == false)
            {
                while(frameBuffer.size() == nBufferCount)
                    ReduceBufferedFrameLifeTime();
                
                startEvent.notify_all();
                frameBufferChangedEvent.wait(lg);
            }
        }
        
        void threadFunction()
        {
            try
            {
                for(;;)
                {
                    boost::this_thread::interruption_point();
                    
                    int nFrameToRead;
                    {
                        unique_lock<recursive_mutex> lg(nextFrameToReadMutex);
                        nFrameToRead = nNextFrameToRead;
                        nNextFrameToRead = GetNextFrameToRead(nNextFrameToRead);
                    }
                    {
                        unique_lock<recursive_mutex> lg(startEventMutex);
                        while(nFrameToRead == -1)
                        {
                            startEvent.wait(lg);
                            {
                                unique_lock<recursive_mutex> lg(nextFrameToReadMutex);
                                nFrameToRead = nNextFrameToRead;
                            }
                        }
                    }
                    PVideoFrame vf = child->GetFrame(nFrameToRead, firstEnv);
                    
                    BufferFrame(nFrameToRead, vf);
                }
            }
            catch(boost::thread_interrupted)
            {
                ;
            }
        }
        
        int GetNextFrameToRead(int nNextFrameToRead)
        {
            ++nNextFrameToRead;
            if(nNextFrameToRead == nFrameCount) return -1;
            else return nNextFrameToRead;
        }
};

AVSValue __cdecl Create_SoraThread(AVSValue args, void* user_data, IScriptEnvironment* env)
{
    if(args[0].IsClip())
    {
        int bufferCount = args[1].IsInt() ? args[1].AsInt() : 5;
        int cacheCount = args[2].IsInt() ? args[2].AsInt() : 10;
        return AVSValue(new SoraThread(args[0].AsClip(), bufferCount, cacheCount));
    }
    else
    {
        env->ThrowError("SoraThread:  No available clip!");
        return AVSValue(0);
    }
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
    env->AddFunction("SoraThread", "c[Buffer]i[Cache]i", Create_SoraThread, 0);
    return 0;
}
